
/*============================================================================
 * antijop, an anti-malware program.
 * Copyright (C) 2014 by Zack T Smith.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * The author may be reached at 1@zsmith.co.
 *===========================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <ctype.h>

#include "defs.h"

static bool found_data_section = false;

#define N_REGS (16)
static char *reg [N_REGS] = {
        "rax", "rbx", "rcx", "rdx", "rsi", "rdi", "rbp", "rsp",
        "r8", "r9", "r10", "r11", "r12", "r13", "r14", "r15",
};
static char *regd [N_REGS] = {
        "eax", "ebx", "ecx", "edx", "esi", "edi", "ebp", "esp",
        "r8d", "r9d", "r10d", "r11d", "r12d", "r13d", "r14d", "r15d",
};
static char *regw [N_REGS] = {
        "ax", "bx", "cx", "dx", "si", "di", "bp", "sp",
        "r8w", "r9w", "r10w", "r11w", "r12w", "r13w", "r14w", "r15w",
};
static char *regb [N_REGS] = {
        "al", "bl", "cl", "dl", "al", "al", "al", "al", // CANNOT USE WITH REX: "ah", "bh", "ch", "dh", 
        "r8b", "r9b", "r10b", "r11b", "r12b", "r13b", "r14b", "r15b",
};
static char *regb2 [N_REGS] = {
        "al", "bl", "cl", "dl", "ah", "bh", "ch", "dh", 
        "r8b", "r9b", "r10b", "r11b", "r12b", "r13b", "r14b", "r15b",
};

void print_only_allowed_ret ()
{
	found_data_section = true;
	puts ("TheOnlyAllowedRET: RET");
}

int 
readline (FILE *f, char *line, int len, char *comment, int commentlen)
{
	if (!f || !line || !comment || len < 2 || commentlen < 2)
		return EOF;

	bzero (line, len);
	bzero (comment, commentlen);

	int ch;
	int ix = 0;
	int commentix = 0;
	while (EOF != (ch = fgetc (f))) {
		if (ch == ';') {
			line [ix] = 0;
			comment [commentix++] = ch;
			while (commentix < commentlen-1 && '\n' != (ch = fgetc (f))) {
				if (ch == EOF) 
					break;
				comment [commentix++] = ch;
			}
			comment [commentix] = 0;
			break;
		}
		if (ch == '\r')
			continue;
		if (ch == '\n')
			break;
		if (isspace (ch)) {
			// RULE: Convert all whitespace into space.
			ch = ' ';
			// RULE: Compress multiple whitespace chars into one space.
			if (ix && line[ix-1] == ' ')
				continue;
		}
		line [ix++] = ch;
		if (ix >= len-1)
			break;
		
	}
	line[ix] = 0;
	if (!ix && !commentix && ch == EOF)
		return EOF;
	return ix;
}

int
getwords (char *line, int len, char **words, int maxwords)
{
	if (!line || !words || len < 1 || maxwords < 1)
		return 0;

	bzero (words, sizeof(char*) * maxwords);

#define IS_TOKEN(XX) ((XX) == '[' || (XX) == ']' || (XX) == '+' || (XX) == '*' || (XX) == ',')
	int ix = 0;
	int n_words = 0;
	while (ix < len && n_words < maxwords) {
		// Skip past whitespace.
		//
		while (ix < len && line[ix] == ' ')
			ix++;
		if (ix >= len)
			break;

		char *start = &line[ix];
		int ch = *start;

		// RULE: Several single characters become 1-char words.
		//
		if (IS_TOKEN(ch)) {
			char str[2] = { ch, 0 };
			words [n_words++] = strdup (str);
			ix++;
			continue;
		}

		// Grab up one entire word and duplicate it.
		//
		int n_chars = 0;
		while (ix < len && !isspace (line[ix])) {
			int ch = line [ix];
			if (ch == '-' && n_chars>0)
				break;
			if (IS_TOKEN(ch)) 
				break;
			ix++;
			n_chars++;
		}
		char *end = &line[ix];
		if (start > end)
			break;
		int wordlen = end-start;
		if (end >= &line[len])
			wordlen++;
		char *word = malloc (wordlen+1);
		if (!word)
			break;
		memcpy (word, start, wordlen);
		word [wordlen] = 0;
		words [n_words++] = word;
//printf ("GOT WORD %s\n", word);
	}
	return n_words;
}

bool is_reg8 (char *str)
{
	if (!str)
		return false;

	int i;
	for (i = 0; i < N_REGS; i++) 
		if (!strcasecmp (str, regb2[i]))
			return true;

	return false;
}

bool is_reg16 (char *str)
{
	if (!str)
		return false;

	int i;
	for (i = 0; i < N_REGS; i++) 
		if (!strcasecmp (str, regw[i]))
			return true;

	return false;
}

bool is_reg32 (char *str)
{
	if (!str)
		return false;

	int i;
	for (i = 0; i < N_REGS; i++) 
		if (!strcasecmp (str, regd[i]))
			return true;

	return false;
}

bool is_reg64 (char *str)
{
	if (!str)
		return false;

	int i;
	for (i = 0; i < N_REGS; i++) 
		if (!strcasecmp (str, reg[i]))
			return true;

	return false;
}

bool
is_register (char *str)
{
	if (!str)
		return false;

	int i;
	for (i = 0; i < N_REGS; i++) 
		if (!strcasecmp (str, reg[i]))
			return true;
	for (i = 0; i < N_REGS; i++) 
		if (!strcasecmp (str, regb[i]))
			return true;
	for (i = 0; i < N_REGS; i++) 
		if (!strcasecmp (str, regw[i]))
			return true;
	for (i = 0; i < N_REGS; i++) 
		if (!strcasecmp (str, regd[i]))
			return true;

	return false;
}

bool contains_bad_byte (uint64_t value)
{
	int i;
	union {
		int64_t value;
		uint64_t uvalue;
		unsigned char bytes[8];
	} v;

	v.uvalue = value;

	for (i = 0; i < 8; i++) {
		if (v.bytes[i] == 0xFF)
			return true;
		if (v.bytes[i] == 0xC3)
			return true;
		if (v.bytes[i] == 0xC2)
			return true;
		if (v.bytes[i] == 0xCA)
			return true;
		if (v.bytes[i] == 0xCB)
			return true;
	}
	return false;
}

uint64_t parse_hex (char *s)
{
	if (!s)
		return 0;

	if ('0' == *s && 'x' == tolower((int) s[1]))
		s += 2;

	unsigned long long value = 0;
	if (1 == sscanf (s, "%llx", &value))
		return (uint64_t) value;

	return 0;
}

bool
convert_to_safe_form (char **words, int n_words)
{
	char *w0 = NULL;
	char *w1 = NULL;
	char *w2 = NULL;
	char *w3 = NULL;
	char *w4 = NULL;
	char *w5 = NULL;
	char *w6 = NULL;
	char *w7 = NULL;
	bool r1 = false;
	bool r2 = false;
	bool r3 = false;

	if (!words || n_words < 1)
		return false;

	// Store any label.
	//
	const char *label = NULL;
	if (words[0][0]) {
		bool found = false;
		if (words[0][0] == '.' && words[0][1] == 'L') {
			label = words[0];
			found = true;
		} else {
			int tmp = strlen (words[0]);
			if (tmp > 0 && ':' == words[0][tmp-1]) {
				label = words[0];
				found = true;
			}
		}
		if (found) {
			int i;
			for (i = 0; i < n_words-1; i++)
				words[i] = words[i+1];
			n_words--;
		}
	}

	if (!n_words && label) {
		return false;
	}

	if (n_words > 0) {
		w0 = words[0];
	}
	if (n_words > 1) {
		w1 = words[1];
		if (is_register (w1))
			r1 = true;
	}
	if (n_words > 2) {
		w2 = words[2];
		if (is_register (w2))
			r2 = true;
	}
	if (n_words > 3) {
		w3 = words[3];
		if (is_register (w3))
			r3 = true;
	}
	if (n_words > 4) {
		w4 = words[4];
	}
	if (n_words > 5) {
		w5 = words[5];
	}
	if (n_words > 6) {
		w6 = words[6];
	}
	if (n_words > 7) {
		w7 = words[7];
	}

	if (n_words == 2 && !strcasecmp ("SECTION", w0) && !strcasecmp (".DATA", w1)) {
		if (!found_data_section) {
			print_only_allowed_ret ();
			if (label) puts (label);	
			printf ("SECTION .DATA");
			return true;
		}
	}

	//--------------------------------------------------
	// RET instruction handling.
	// RULE: All RETs converted into JMP to final RET.
	//
	if (n_words == 1 && !strcasecmp ("RET", w0)) {
		if (label) puts (label);	
		puts ("JMP TheOnlyAllowedRET");
		return true;
	}

	//-----------------------------------------------------
	// MOV rdi, rdi (contains 0xff) and other moves between
	// exact same register.
	//
	if (n_words == 4 && !strcasecmp (w0, "MOV") && is_register(w1) && !strcasecmp (w1, w3) && *w2==',') {
		if (label) puts (label);
		printf ("NOP");
		return true;
	}

	//------------------------------
	// MOV rdi, r15 contains 0xff.
	// MOV r15, rdi contains 0xff.
	//
	if (n_words == 4 && !strcasecmp (w0, "MOV") && !strcasecmp(w1, "rdi") && !strcasecmp (w3, "r15") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH r15");
		printf ("POP rdi");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "MOV") && !strcasecmp(w3, "rdi") && !strcasecmp (w1, "r15") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		printf ("POP r15");
		return true;
	}

	//------------------------------
	// MOV bl, al contains 0xC3.
	// MOV bx, ax contains 0xC3.
	// MOV ebx, eax contains 0xC3.
	// MOV rbx, rax contains 0xC3.
	//
	if (n_words == 4 && !strcasecmp (w0, "MOV") && !strcasecmp(w1, "bl") && !strcasecmp (w3, "al") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH ax");
		puts ("MOV bl, byte [rsp]");
		printf ("LEA rsp, [rsp+2]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "MOV") && !strcasecmp(w1, "bx") && !strcasecmp (w3, "ax") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH ax");
		printf ("POP bx");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "MOV") && !strcasecmp(w1, "ebx") && !strcasecmp (w3, "eax") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rax");
		puts ("MOV ebx, dword [rsp]");
		printf ("LEA rsp, [rsp+8]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "MOV") && !strcasecmp(w1, "rbx") && !strcasecmp (w3, "rax") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rax");
		printf ("POP rbx");
		return true;
	}

	//------------------------------
	// MOV dl, al contains 0xC2.
	// MOV dx, ax contains 0xC2.
	// MOV edx, eax contains 0xC2.
	// MOV rdx, rax contains 0xC2.
	//
	if (n_words == 4 && !strcasecmp (w0, "MOV") && !strcasecmp(w1, "dl") && !strcasecmp (w3, "al") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH ax");
		puts ("MOV dl, byte [rsp]");
		printf ("LEA rsp, [rsp+2]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "MOV") && !strcasecmp(w1, "dx") && !strcasecmp (w3, "ax") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH ax");
		printf ("POP dx");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "MOV") && !strcasecmp(w1, "edx") && !strcasecmp (w3, "eax") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rax");
		puts ("MOV edx, dword [rsp]");
		printf ("LEA rsp, [rsp+8]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "MOV") && !strcasecmp(w1, "rdx") && !strcasecmp (w3, "rax") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rax");
		printf ("POP rdx");
		return true;
	}

	//------------------------------
	// XOR rdi, rdi contains 0xff.
	// XOR r15, r15 contains 0xff.
	// XOR r15, rdi contains 0xff.
	// XOR rdi, r15 contains 0xff.
	//
	if (n_words == 4 && !strcasecmp (w0, "XOR") && !strcasecmp(w1, "rdi") && !strcasecmp (w3, "rdi") && *w2==',') {
		if (label) puts (label);
		printf ("AND rdi, BYTE 0");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "XOR") && !strcasecmp(w1, "r15") && !strcasecmp (w3, "r15") && *w2==',') {
		if (label) puts (label);
		printf ("AND r15, BYTE 0");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "XOR") && !strcasecmp(w1, "rdi") && !strcasecmp (w3, "r15") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH r15");
		puts ("XOR rdi, qword [rsp]");
		printf ("POP r15");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "XOR") && !strcasecmp(w1, "r15") && !strcasecmp (w3, "rdi") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("XOR r15, qword [rsp]");
		printf ("POP rdi");
		return true;
	}

	//------------------------------
	// ADD rdi, rdi contains 0xff.
	// ADD r15, r15 contains 0xff.
	// ADD r15, rdi contains 0xff.
	// ADD rdi, r15 contains 0xff.
	//
	if (n_words == 4 && !strcasecmp (w0, "ADD") && !strcasecmp(w1, "rdi") && !strcasecmp (w3, "rdi") && *w2==',') {
		if (label) puts (label);
		// printf ("SHL rdi, 1");
		puts ("PUSH rdi");
		puts ("ADD rdi, qword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "ADD") && !strcasecmp(w1, "r15") && !strcasecmp (w3, "r15") && *w2==',') {
		if (label) puts (label);
		// printf ("SHL r15, 1");
		puts ("PUSH r15");
		puts ("ADD r15, qword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "ADD") && !strcasecmp(w1, "rdi") && !strcasecmp (w3, "r15") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH r15");
		puts ("ADD rdi, qword [rsp]");
		printf ("POP r15");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "ADD") && !strcasecmp(w1, "r15") && !strcasecmp (w3, "rdi") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("ADD r15, qword [rsp]");
		printf ("POP rdi");
		return true;
	}

	//------------------------------
	// SUB rdi, rdi contains 0xff.
	// SUB r15, r15 contains 0xff.
	// SUB r15, rdi contains 0xff.
	// SUB rdi, r15 contains 0xff.
	//
	if (n_words == 4 && !strcasecmp (w0, "SUB") && !strcasecmp(w1, "rdi") && !strcasecmp (w3, "rdi") && *w2==',') {
		if (label) puts (label);
		// printf ("SHL rdi, 64");
		puts ("PUSH rdi");
		puts ("SUB rdi, qword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "SUB") && !strcasecmp(w1, "r15") && !strcasecmp (w3, "r15") && *w2==',') {
		if (label) puts (label);
		// printf ("SHL r15, 64");
		puts ("PUSH r15");
		puts ("SUB r15, qword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "SUB") && !strcasecmp(w1, "rdi") && !strcasecmp (w3, "r15") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH r15");
		puts ("SUB rdi, qword [rsp]");
		printf ("POP r15");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "SUB") && !strcasecmp(w1, "r15") && !strcasecmp (w3, "rdi") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("SUB r15, qword [rsp]");
		printf ("POP rdi");
		return true;
	}

	//------------------------------
	// ADC rdi, rdi contains 0xff.
	// ADC r15, r15 contains 0xff.
	// ADC r15, rdi contains 0xff.
	// ADC rdi, r15 contains 0xff.
	//
	if (n_words == 4 && !strcasecmp (w0, "ADC") && !strcasecmp(w1, "rdi") && !strcasecmp (w3, "rdi") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("ADC rdi, qword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "ADC") && !strcasecmp(w1, "r15") && !strcasecmp (w3, "r15") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH r15");
		puts ("ADC r15, qword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "ADC") && !strcasecmp(w1, "rdi") && !strcasecmp (w3, "r15") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH r15");
		puts ("ADC rdi, qword [rsp]");
		printf ("POP r15");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "ADC") && !strcasecmp(w1, "r15") && !strcasecmp (w3, "rdi") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("ADC r15, qword [rsp]");
		printf ("POP rdi");
		return true;
	}

	//------------------------------
	// SBB rdi, rdi contains 0xff.
	// SBB r15, r15 contains 0xff.
	// SBB r15, rdi contains 0xff.
	// SBB rdi, r15 contains 0xff.
	//
	if (n_words == 4 && !strcasecmp (w0, "SBB") && !strcasecmp(w1, "rdi") && !strcasecmp (w3, "rdi") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("SBB rdi, qword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "SBB") && !strcasecmp(w1, "r15") && !strcasecmp (w3, "r15") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH r15");
		puts ("SBB r15, qword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "SBB") && !strcasecmp(w1, "rdi") && !strcasecmp (w3, "r15") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH r15");
		puts ("SBB rdi, qword [rsp]");
		puts ("POP r15");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "SBB") && !strcasecmp(w1, "r15") && !strcasecmp (w3, "rdi") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("SBB r15, qword [rsp]");
		puts ("POP rdi");
		return true;
	}

	//------------------------------
	// AND rdi, rdi contains 0xff.
	// AND r15, r15 contains 0xff.
	// AND r15, rdi contains 0xff.
	// AND rdi, r15 contains 0xff.
	//
	if (n_words == 4 && !strcasecmp (w0, "AND") && !strcasecmp(w1, "rdi") && !strcasecmp (w3, "rdi") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("AND rdi, qword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "AND") && !strcasecmp(w1, "r15") && !strcasecmp (w3, "r15") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH r15");
		puts ("AND r15, qword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "AND") && !strcasecmp(w1, "rdi") && !strcasecmp (w3, "r15") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH r15");
		puts ("AND rdi, qword [rsp]");
		printf ("POP r15");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "AND") && !strcasecmp(w1, "r15") && !strcasecmp (w3, "rdi") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("AND r15, qword [rsp]");
		printf ("POP rdi");
		return true;
	}

	//------------------------------
	// OR rdi, rdi contains 0xff.
	// OR r15, r15 contains 0xff.
	// OR r15, rdi contains 0xff.
	// OR rdi, r15 contains 0xff.
	//
	if (n_words == 4 && !strcasecmp (w0, "OR") && !strcasecmp(w1, "rdi") && !strcasecmp (w3, "rdi") && *w2==',') {
		if (label) puts (label);
		// OR rdi,rdi is presumed to be intended to set the Z flag.
		puts ("PUSH rdi");
		puts ("OR rdi, qword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "OR") && !strcasecmp(w1, "r15") && !strcasecmp (w3, "r15") && *w2==',') {
		if (label) puts (label);
		// OR r15,r15 is presumed to be intended to set the Z flag.
		puts ("PUSH r15");
		puts ("OR r15, qword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "OR") && !strcasecmp(w1, "rdi") && !strcasecmp (w3, "r15") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH r15");
		puts ("OR rdi, qword [rsp]");
		printf ("POP r15");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "OR") && !strcasecmp(w1, "r15") && !strcasecmp (w3, "rdi") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("OR r15, qword [rsp]");
		printf ("POP rdi");
		return true;
	}

	//------------------------------
	// TEST rdi, rdi contains 0xff.
	// TEST r15, r15 contains 0xff.
	// TEST r15, rdi contains 0xff.
	// TEST rdi, r15 contains 0xff.
	//
	if (n_words == 4 && !strcasecmp (w0, "TEST") && !strcasecmp(w1, "rdi") && !strcasecmp (w3, "rdi") && *w2==',') {
		if (label) puts (label);
		// TEST rdi,rdi is presumed to be intended to set the Z flag.
		puts ("PUSH rdi");
		puts ("TEST rdi, qword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "TEST") && !strcasecmp(w1, "r15") && !strcasecmp (w3, "r15") && *w2==',') {
		if (label) puts (label);
		// TEST r15,r15 is presumed to be intended to set the Z flag.
		puts ("PUSH r15");
		puts ("TEST r15, qword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "TEST") && !strcasecmp(w1, "rdi") && !strcasecmp (w3, "r15") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH r15");
		puts ("TEST rdi, qword [rsp]");
		printf ("POP r15");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "TEST") && !strcasecmp(w1, "r15") && !strcasecmp (w3, "rdi") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("TEST r15, qword [rsp]");
		printf ("POP rdi");
		return true;
	}

	//------------------------------
	// CMP rdi, rdi contains 0xff.
	// CMP r15, r15 contains 0xff.
	// CMP r15, rdi contains 0xff.
	// CMP rdi, r15 contains 0xff.
	//
	if (n_words == 4 && !strcasecmp (w0, "CMP") && !strcasecmp(w1, "rdi") && !strcasecmp (w3, "rdi") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("CMP rdi, qword [rsp]");
		printf ("POP rdi");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "CMP") && !strcasecmp(w1, "r15") && !strcasecmp (w3, "r15") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH r15");
		puts ("CMP r15, qword [rsp]");
		printf ("POP r15");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "CMP") && !strcasecmp(w1, "rdi") && !strcasecmp (w3, "r15") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH r15");
		puts ("CMP rdi, qword [rsp]");
		printf ("POP r15");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "CMP") && !strcasecmp(w1, "r15") && !strcasecmp (w3, "rdi") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("CMP r15, qword [rsp]");
		printf ("POP rdi");
		return true;
	}

	//------------------------------
	// ADD edi, edi contains 0xff.
	// SUB edi, edi contains 0xff.
	// ADC edi, edi contains 0xff.
	// SBB edi, edi contains 0xff.
	// AND edi, edi contains 0xff.
	// OR  edi, edi contains 0xff.
	// XOR edi, edi contains 0xff.
	// TEST edi, edi contains 0xff.
	//
	if (n_words == 4 && !strcasecmp (w0, "ADD") && !strcasecmp(w1, "edi") && !strcasecmp (w3, "edi") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("ADD edi, dword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "SUB") && !strcasecmp(w1, "edi") && !strcasecmp (w3, "edi") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("SUB edi, dword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "ADC") && !strcasecmp(w1, "edi") && !strcasecmp (w3, "edi") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("ADC edi, dword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "SBB") && !strcasecmp(w1, "edi") && !strcasecmp (w3, "edi") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("SBB edi, dword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "AND") && !strcasecmp(w1, "edi") && !strcasecmp (w3, "edi") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("AND edi, dword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "OR") && !strcasecmp(w1, "edi") && !strcasecmp (w3, "edi") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("OR edi, dword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "XOR") && !strcasecmp(w1, "edi") && !strcasecmp (w3, "edi") && *w2==',') {
		// NOTE: We cannot use a simple MOV edi, 0 because setting of the Z bit may be important
		// to subsequent code.
		if (label) puts (label);
		puts ("PUSH rdi");
		puts ("XOR edi, dword [rsp]");
		printf ("LEA rsp, [8+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "TEST") && !strcasecmp(w1, "edi") && !strcasecmp (w3, "edi") && *w2==',') {
		if (label) puts (label);
		puts ("OR edi, dword 0");
		return true;
	}

	//------------------------------
	// ADD di, di contains 0xff.
	// SUB di, di contains 0xff.
	// ADC di, di contains 0xff.
	// SBB di, di contains 0xff.
	// AND di, di contains 0xff.
	// OR  di, di contains 0xff.
	// XOR di, di contains 0xff.
	// TEST di, di contains 0xff.
	//
	if (n_words == 4 && !strcasecmp (w0, "ADD") && !strcasecmp(w1, "di") && !strcasecmp (w3, "di") && *w2==',') {
		if (label) puts (label);
		printf ("SHL di, 1");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "SUB") && !strcasecmp(w1, "di") && !strcasecmp (w3, "di") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH di");
		puts ("SUB di, word [rsp]");
		printf ("LEA rsp, [2+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "ADC") && !strcasecmp(w1, "di") && !strcasecmp (w3, "di") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH di");
		puts ("ADC di, word [rsp]");
		printf ("LEA rsp, [2+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "SBB") && !strcasecmp(w1, "di") && !strcasecmp (w3, "di") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH di");
		puts ("SBB di, word [rsp]");
		printf ("LEA rsp, [2+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "AND") && !strcasecmp(w1, "di") && !strcasecmp (w3, "di") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH di");
		puts ("AND di, word [rsp]");
		printf ("LEA rsp, [2+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "OR") && !strcasecmp(w1, "di") && !strcasecmp (w3, "di") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH di");
		puts ("OR di, word [rsp]");
		printf ("LEA rsp, [2+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "XOR") && !strcasecmp(w1, "di") && !strcasecmp (w3, "di") && *w2==',') {
		if (label) puts (label);
		puts ("PUSH di");
		puts ("XOR di, word [rsp]");
		printf ("LEA rsp, [2+rsp]");
		return true;
	}
	if (n_words == 4 && !strcasecmp (w0, "TEST") && !strcasecmp(w1, "di") && !strcasecmp (w3, "di") && *w2==',') {
		if (label) puts (label);
		printf ("OR di, word 0");
		return true;
	}

	//------------------------------
	// INC bl contains 0xC3.
	// INC dl contains 0xC2.
	// INC bx contains 0xC3.
	// INC dx contains 0xC2.
	// INC ebx contains 0xC3.
	// INC edx contains 0xC2.
	// INC rbx contains 0xC3.
	// INC rdx contains 0xC2.
	//
	if (n_words == 2 && !strcasecmp (w0, "INC") && 
		(!strcasecmp ("bl",w1) || !strcasecmp ("dl",w1) ||
		!strcasecmp ("bx",w1) || !strcasecmp ("dx",w1) ||
		!strcasecmp ("ebx",w1) || !strcasecmp ("edx",w1) ||
		!strcasecmp ("rbx",w1) || !strcasecmp ("rbx",w1)
	)) {
		if (label) puts (label);
		// Unfortunately ADD bl,1 = 80 c3 01
		// and ADD dl,1 = 80 c2 01
		puts ("PUSHF");
		puts ("CLC");
		printf ("ADC %s, byte 1\n", w1);
		puts ("POPF"); // Z flag lost here while restoring carry.
		printf ("TEST %s, %s", w1, w1); // Z flag reestablished.
		return true;
	}

	//------------------------------
	// INC reg contains 0xff.
	//
	if (n_words == 2 && !strcasecmp (w0, "INC") && is_register (w1) && !is_reg8 (w1)) {
		if (label) puts (label);
		if (is_reg64 (w1)) {
			printf ("LEA %s, [1+%s]\n", w1, w1); // Do the increment.

			if (strcasecmp ("rdi", w1)) {
				printf ("OR %s, %s\n", w1, w1); // Set the Z flag.
			} else {
				puts ("PUSH rdi");
				puts ("TEST qword [rsp], 0");
				printf ("LEA rsp, [8+rsp]");
			}
		} else {
			printf ("ADD %s, 1", w1); // Do the increment.
			// XX Problem: carry flag corrupted. Alternative is PUSHF, ADD, POPF, TEST.
		}
		return true;
	}
	//------------------------------
	// DEC reg contains 0xff.
	//
	if (n_words == 2 && !strcasecmp (w0, "DEC") && is_register (w1)) {
		if (label) puts (label);
		printf ("SUB %s, 1", w1);
		// XX Small problem, carry flag gets affected.
		return true;
	}
	//------------------------------
	// INC qword [reg] contains 0xff.
	// INC dword [reg] contains 0xff.
	// INC word [reg] contains 0xff.
	// INC byte [reg] contains 0xfe (OK).
	//
	if (n_words == 5 && !strcasecmp (w0, "INC") && !strcasecmp ("qword",w1) && '[' == *w2 && is_register (w3) && ']' == *w4) {
		if (label) puts (label);
		printf ("ADD qword [%s], 1", w3);
		// XX Small problem, carry flag gets affected.
		return true;
	}
	if (n_words == 5 && !strcasecmp (w0, "INC") && !strcasecmp ("dword",w1) && '[' == *w2 && is_register (w3) && ']' == *w4) {
		if (label) puts (label);
		printf ("ADD dword [%s], 1", w3);
		// XX Small problem, carry flag gets affected.
		return true;
	}
	if (n_words == 5 && !strcasecmp (w0, "INC") && !strcasecmp ("word",w1) && '[' == *w2 && is_register (w3) && ']' == *w4) {
		if (label) puts (label);
		printf ("ADD word [%s], 1", w3);
		// XX Small problem, carry flag gets affected.
		return true;
	}

	//------------------------------
	// IMUL edi,edi has 0xFF
	// IMUL rdi,rdi has 0xFF
	// IMUL r15,r15 has 0xFF
	// IMUL r15d,r15d has 0xFF
	//
	if (n_words == 4 && !strcasecmp (w0, "IMUL") && (
		!strcasecmp ("di",w1)  ||
		!strcasecmp ("edi",w1)  ||
		!strcasecmp ("rdi",w1)  ||
		!strcasecmp ("r15w",w1)  ||
		!strcasecmp ("r15d",w1)  ||
		!strcasecmp ("r15",w1) 
	    ) && ',' == *w2 && !strcasecmp (w1, w3)) 
	{
		if (label) puts (label);
		puts ("PUSH RCX");
		printf ("MOV rcx, %s\n", w1);
		printf ("IMUL rcx, rcx\n");
		printf ("MOV %s, rcx\n", w1);
		puts ("POP RCX");
		return true;
	}

	//------------------------------
	// FFREE ST3 contains 0xC3.
	//
	if (n_words == 2 && !strcasecmp (w0, "FFREE")) {
		// TO DO
	}

	//------------------------------
	// FCOS contains 0xFF.
	//
	if (n_words == 1 && !strcasecmp (w0, "FCOS")) {
		if (label) puts (label);
		// cos(x) = sin(x + pi/2)
		puts ("FLDPI");
		puts ("FILD 2");
		puts ("FDIV");
		printf ("FSIN"); 
		// XX Problem if stack overflows.
		return true;
	}

	//----------------------------------------
	// These contain 0xC3/0xC2 in MOD/RM.
	// ADD bl/dl, immediate 
	// ADD bx/dx, immediate 
	// ADD ebx/edx, immediate 
	// ADD rbx/rdx, immediate 
	//
	if (n_words == 4 && !strcasecmp (w0, "ADD") && (
		!strcasecmp ("bl",w1) || !strcasecmp ("dl",w1) ||
		!strcasecmp ("bx",w1) || !strcasecmp ("dx",w1) ||
		!strcasecmp ("ebx",w1) || !strcasecmp ("edx",w1) ||
		!strcasecmp ("rbx",w1) || !strcasecmp ("rbx",w1)
	    )  
	    && ','==*w2 && (isdigit(*w3) || '-' == *w3)) 
	{
		union {
			int64_t value;
			uint64_t uvalue;
			unsigned char bytes[8];
		} v;
		if ('0' == *w3 && 'x' == w3[1]) {
			v.value = parse_hex (w3);
		} else {
			v.value = atoi (w3);
		}

		// NOTE: While ADD is problematic, ADC is not.

		if (label) puts (label);
		puts ("CLC");

		if (!contains_bad_byte (v.uvalue)) {
			printf ("ADC %s, 0x%llx\n", w1, v.uvalue);
		} else {
			// RULE: If this addition's immediate has bad byte(s),
			// break the immediate into even and odd bits.

			printf ("ADC %s, 0x%llx\n", w1, v.uvalue & 0xaaaaaaaaaaaaaaaa);
			puts ("CLC");
			printf ("ADC %s, 0x%llx", w1, v.uvalue & 0x5555555555555555);
		}

		return true;
	}

	//----------------------------------------
	// These contain 0xC3/0xC2 in MOD/RM.
	// TEST bl/dl, immediate 
	// TEST bx/dx, immediate 
	// TEST ebx/edx, immediate 
	// TEST rbx/rdx, immediate 
	//
	if (n_words == 4 && !strcasecmp (w0, "TEST") && (
		!strcasecmp ("bl",w1) || !strcasecmp ("dl",w1) ||
		!strcasecmp ("bx",w1) || !strcasecmp ("dx",w1) ||
		!strcasecmp ("ebx",w1) || !strcasecmp ("edx",w1) ||
		!strcasecmp ("rbx",w1) || !strcasecmp ("rbx",w1)
	    )  
	    && ','==*w2 && (isdigit(*w3) || '-' == *w3)) 
	{
		bool using_rbx = false;
		if (!strcasecmp ("bl",w1) || !strcasecmp ("bx",w1) || !strcasecmp ("ebx",w1) || !strcasecmp ("rbx",w1)) 
			using_rbx = true;

		union {
			int64_t value;
			uint64_t uvalue;
			unsigned char bytes[8];
		} v;
		if ('0' == *w3 && 'x' == w3[1]) {
			v.value = parse_hex (w3);
		} else {
			v.value = atoi (w3);
		}

		if (label) puts (label);
		if (using_rbx) {
			puts ("PUSH rbx");
		} else {
			puts ("PUSH rdx");
		}
		if (is_reg8 (w1)) printf ("TEST byte [rsp], %s\n", w3);
		else if (is_reg16 (w1)) printf ("TEST word [rsp], %s\n", w3);
		else if (is_reg32 (w1)) printf ("TEST dword [rsp], %s\n", w3);
		else if (is_reg64 (w1)) printf ("TEST qword [rsp], %s\n", w3);
		printf ("LEA rsp, [8+rsp]");
		return true;
	}

	//----------------------------------------
	// These contain 0xC3/0xC2 in MOD/RM.
	// OR bl/dl, immediate 
	// OR bx/dx, immediate 
	// OR ebx/edx, immediate 
	// OR rbx/rdx, immediate 
	//
	if (n_words == 4 && !strcasecmp (w0, "OR") && (
		!strcasecmp ("bl",w1) || !strcasecmp ("dl",w1) ||
		!strcasecmp ("bx",w1) || !strcasecmp ("dx",w1) ||
		!strcasecmp ("ebx",w1) || !strcasecmp ("edx",w1) ||
		!strcasecmp ("rbx",w1) || !strcasecmp ("rbx",w1)
	    )  
	    && ','==*w2 && (isdigit(*w3) || '-' == *w3)) 
	{
		bool using_rbx = false;
		if (!strcasecmp ("bl",w1) || !strcasecmp ("bx",w1) || !strcasecmp ("ebx",w1) || !strcasecmp ("rbx",w1)) 
			using_rbx = true;

		if (label) puts (label);
		if (using_rbx) {
			puts ("PUSH rbx");
		} else {
			puts ("PUSH rdx");
		}

		if (is_reg8 (w1)) {
			printf ("OR byte [rsp], %s\n", w3);
		}
		else if (is_reg16 (w1)) {
			printf ("OR word [rsp], %s\n", w3);
		}
		else if (is_reg32 (w1)) {
			printf ("OR dword [rsp], %s\n", w3);
		}
		else if (is_reg64 (w1)) {
			printf ("OR qword [rsp], %s\n", w3);
		}

		printf ("MOV %s, [rsp]\n", w1);
		printf ("LEA rsp, [8+rsp]");
		return true;
	}

	//----------------------------------------
	// These contain 0xC3/0xC2 in MOD/RM.
	// MOV bl/dl, immediate 
	// MOV bx/dx, immediate 
	// MOV ebx/edx, immediate 
	// MOV rbx/rdx, immediate 
	//
	if (n_words == 4 && !strcasecmp (w0, "MOV") && (
		!strcasecmp ("bl",w1) || !strcasecmp ("dl",w1) ||
		!strcasecmp ("bx",w1) || !strcasecmp ("dx",w1) ||
		!strcasecmp ("ebx",w1) || !strcasecmp ("edx",w1) ||
		!strcasecmp ("rbx",w1) || !strcasecmp ("rbx",w1)
	    )  
	    && ','==*w2 && (isdigit(*w3) || '-' == *w3)) 
	{
		union {
			int64_t value;
			uint64_t uvalue;
			unsigned char bytes[8];
		} v;
		if ('0' == *w3 && 'x' == w3[1]) {
			v.value = parse_hex (w3);
		} else {
			v.value = atoi (w3);
		}

		bool using_rbx = false;
		if (!strcasecmp ("bl",w1) || !strcasecmp ("bx",w1) || !strcasecmp ("ebx",w1) || !strcasecmp ("rbx",w1)) 
			using_rbx = true;

		if (label) puts (label);
		puts ("PUSHF"); 
		printf ("XOR %s, %s\n", w1, w1);

		if (!contains_bad_byte (v.uvalue)) {
			printf ("XOR %s, %s\n", w1, w3);
		} else {
			// RULE: If immediate contains 0xFF or 0xC3 then break it into even & odd bits.

			printf ("XOR %s, 0x%llx\n", w1, v.uvalue & 0xaaaaaaaaaaaaaaaa);
			printf ("XOR %s, 0x%llx\n", w1, v.uvalue & 0x5555555555555555);
		}
		puts ("POPF"); // Restore Z flag etc.

		return true;
	}

	//----------------------------------------
	// MOV reg1, immediate
	// where immediate contains 0xC3 or 0xC3.
	// NOTE: This addresses ROP gadgets.
	//
	if (n_words == 4 && !strcasecmp (w0, "MOV") && is_register (w1) && ','==*w2 && (isdigit(*w3) || '-' == *w3)) {
		union {
			int64_t value;
			uint64_t uvalue;
			unsigned char bytes[8];
		} v;
		if ('0' == *w3 && 'x' == w3[1]) {
			v.value = parse_hex (w3);
		} else {
			v.value = atoi (w3);
		}

		//----------------------------------------
		// RULE: If any byte as 0xC3/0xFF, break 
		// it apart into even/odd bits.
		//
		if (contains_bad_byte (v.uvalue)) {
			uint64_t value = v.uvalue;
			if (is_reg64(w1)) {
				if (label) puts (label);
				puts ("PUSHF"); 
				printf ("MOV %s, 0x%llx\n", w1, value & 0xaaaaaaaaaaaaaaaa);
				printf ("XOR %s, 0x%llx\n", w1, value & 0x5555555555555555);
				puts ("POPF"); 
				return true;
			}
			if (is_reg32(w1)) {
				if (label) puts (label);
				puts ("PUSHF"); 
				printf ("MOV %s, 0x%lx\n", w1, (unsigned long) (0xaaaaaaaa & value));
				printf ("XOR %s, 0x%lx\n", w1, (unsigned long) (0x55555555 & value));
				puts ("POPF"); 
				return true;
			}
			if (is_reg16(w1)) {
				if (label) puts (label);
				puts ("PUSHF"); 
				printf ("MOV %s, 0x%x\n", w1, (unsigned) (0xaaaa & value));
				printf ("XOR %s, 0x%x\n", w1, (unsigned) (0x5555 & value));
				puts ("POPF"); 
				return true;
			}
			if (is_reg8(w1)) {
				if (label) puts (label);
				puts ("PUSHF"); 
				printf ("MOV %s, 0x%x\n", w1, (unsigned) (0xaa & value));
				printf ("XOR %s, 0x%x\n", w1, (unsigned) (0x55 & value));
				puts ("POPF"); 
				return true;
			}
		} else {
			// RULE: If there are no bad bytes, let the MOV pass through.
		}
	}

	//------------------------------
	// MOV reg1, immediate
	// ... where immediate contains 0xFF.
	//
	if (n_words == 4 && !strcasecmp (w0, "MOV") && is_register (w1) && ','==*w2 && (isdigit(*w3) || '-' == *w3)) {
		union {
			int64_t value;
			uint64_t uvalue;
			unsigned char bytes[8];
		} v;
		if ('0' == *w3 && 'x' == w3[1]) {
			v.value = parse_hex (w3);
		} else {
			v.value = atoi (w3);
		}

		// RULE: If immediate is 0, try to use XOR.
		if (!v.uvalue) {
			if (strcasecmp (w1, "di") && strcasecmp (w1, "edi") && strcasecmp (w1, "rdi")
			 && strcasecmp (w1, "r15") && strcasecmp (w1, "r15d") && strcasecmp (w1, "r15w")) {
				if (label) puts (label);
				puts ("PUSHF"); 
				printf ("XOR %s, %s\n", w1, w1);
				puts ("POPF"); 
				return true;
			}
		}

		//----------------------------------------
		// RULE: If it's a pure 0xff, 0xffff etc
		// just use a XOR+NOT.
		// NOTE: XOR EDI,EDI etc have 0xff in them.
		//
		if (v.value == -1 &&
			strcasecmp (w1, "di") && strcasecmp (w1, "edi") && strcasecmp (w1, "rdi")
			 && strcasecmp (w1, "r15") && strcasecmp (w1, "r15d") && strcasecmp (w1, "r15w"))
		{
			if (label) puts (label);
			puts ("PUSHF"); 
			printf ("XOR %s, %s\n", w1, w1);
			printf ("NOT %s\n", w1);
			puts ("POPF"); 
			return true;
		}
		
		//----------------------------------------
		// RULE: If none of the bytes is zero, 
		// just use a MOV+NOT.
		// XX Caveat: Z bit gets changed.
		//
		if (contains_bad_byte (v.uvalue)) {
			if (is_reg64(w1) && (v.bytes[0] && v.bytes[1] && v.bytes[2] && v.bytes[3] && v.bytes[4] && v.bytes[5] && v.bytes[6] && v.bytes[7])) {
				if (label) puts (label);
				v.uvalue ^= 0xffffffffffffffff;
				printf ("MOV %s, 0x%llx\n", w1, v.value);
				puts ("PUSHF"); 
				printf ("NOT %s\n", w1);
				puts ("POPF"); 
				return true;
			}
			if (is_reg32(w1) && (v.bytes[0] && v.bytes[1] && v.bytes[2] && v.bytes[3])) {
				if (label) puts (label);
				v.uvalue ^= 0xffffffff;
				printf ("MOV %s, 0x%lx\n", w1, (unsigned long) (0xffffffff & v.value));
				puts ("PUSHF"); 
				printf ("NOT %s\n", w1);
				puts ("POPF"); 
				return true;
			}
			if (is_reg16(w1) && (v.bytes[0] && v.bytes[1])) {
				if (label) puts (label);
				v.uvalue ^= 0xffff;
				printf ("MOV %s, 0x%x\n", w1, (unsigned) (0xffff & v.value));
				puts ("PUSHF"); 
				printf ("NOT %s\n", w1);
				puts ("POPF"); 
				return true;
			}
			if (is_reg8(w1) && v.bytes[0]) {
				if (label) puts (label);
				v.uvalue ^= 0xff;
				printf ("MOV %s, 0x%x\n", w1, (unsigned) (0xff & v.value));
				puts ("PUSHF"); 
				printf ("NOT %s\n", w1);
				puts ("POPF"); 
				return true;
			}
		}

		//----------------------------------------
		// RULE: If there's no clean way, break up 
		// the bits and use OR.
		// XX Caveat: Z bit gets changed.
		//
		if (contains_bad_byte (v.uvalue)) {
			uint64_t value = v.uvalue;
			if (is_reg64(w1)) {
				v.value ^= -1;
				if (label) puts (label);
				printf ("MOV %s, %llu\n", w1, value & 0xaaaaaaaaaaaaaaaa);
				printf ("OR %s, %llu\n", w1, value & 0x5555555555555555);
				return true;
			}
			if (is_reg32(w1)) {
				v.value ^= -1;
				if (label) puts (label);
				printf ("MOV %s, %lu\n", w1, (unsigned long) (0xaaaaaaaa & value));
				printf ("OR %s, %lu\n", w1, (unsigned long) (0x55555555 & value));
				return true;
			}
			if (is_reg16(w1)) {
				v.value ^= -1;
				if (label) puts (label);
				printf ("MOV %s, %u\n", w1, (unsigned) (0xaaaa & value));
				printf ("OR %s, %u\n", w1, (unsigned) (0x5555 & value));
				return true;
			}
			if (is_reg8(w1)) {
				v.value ^= -1;
				if (label) puts (label);
				printf ("MOV %s, %u\n", w1, (unsigned) (0xaa & value));
				printf ("OR %s, %u\n", w1, (unsigned) (0x55 & value));
				return true;
			}
		}
	}

	//**************************************************
	// CAVEAT: Need a proper effective address parser.
	// 	What follows below is a quickie solution.
	//**************************************************

	// TO DO
	// NOTE: effective address EBX + EAX*8 contains 0xC3.
	// NOTE: effective address EDX + EAX*8 contains 0xC2.
	// NOTE: effective address EDI + EDI*8 contains 0xFF.

	//------------------------------
	// MOV reg1, [ -1 + reg2 ]
	// ... where reg2 != RSP.
	// NOTE: This is an unlikely case since the source would not be 8-byte aligned but the destination is 64-bits..
	//
	if (n_words == 8 && !strcasecmp (w0, "MOV") && is_reg64 (w1) && ',' == *w2 && '[' == *w3 && !strcmp ("-1", w4) && '+' == *w5 && is_reg64 (w6) && ']' == *w7) {
		if (strcasecmp (w6, "RSP")) {
			char *reg1 = w1;
			char *reg2 = w6;
			if (label) puts (label);
			printf ("MOV %s, %s\n", reg1, reg2);
			printf ("SUB %s, 1\n", reg2);
			printf ("MOV %s, [%s]\n", reg1, reg1);
			return true;
		}
	}

	//------------------------------
	// ADD reg1, [ -1 + reg2 ]
	// ... where reg1 is a byte register.
	// Example: 02 60 ff = ADD AL,[-1 + RAX]
	//
	bool match = !strcasecmp (w0, "ADD") || !strcasecmp (w0, "SUB") || !strcasecmp (w0, "MOV") || !strcasecmp (w0, "OR")
		|| !strcasecmp (w0, "AND") || !strcasecmp (w0, "XOR") || !strcasecmp (w0, "TEST");
	if (n_words == 8 && match && is_reg8 (w1) && ',' == *w2 && '[' == *w3 && !strcmp ("-1", w4) && '+' == *w5 && is_register (w6) && ']' == *w7) {
		char *reg1 = w1;
		char *reg2 = w6;
		if (label) puts (label);
		printf ("SUB %s, 1\n", reg2);
		printf ("%s %s, [%s]\n", w0, reg1, reg2);
		printf ("LEA %s,[%s+1]\n", reg2, reg2);
		return true;
	}
	// ADD reg1, [ reg2 - 1]
	// ... where reg1 is a byte register.
	if (n_words == 8 && match && is_reg8 (w1) && ',' == *w2 && '[' == *w3 && is_register(w4) && '-' == *w5 && !strcmp("1",w6) && ']' == *w7) {
		char *reg1 = w1;
		char *reg2 = w4;
		if (label) puts (label);
		printf ("SUB %s, 1\n", reg2);
		printf ("%s %s, [%s]\n", w0, reg1, reg2);
		printf ("LEA %s,[%s+1]\n", reg2, reg2);
		return true;
	}
	// ADD reg1, [ reg2 -1]
	// ... where reg1 is a byte register.
	if (n_words == 7 && match && is_reg8 (w1) && ',' == *w2 && '[' == *w3 && is_register(w4) && !strcmp("-1",w5) && ']' == *w6) {
		char *reg1 = w1;
		char *reg2 = w4;
		if (label) puts (label);
		printf ("SUB %s, 1\n", reg2);
		printf ("%s %s, [%s]\n", w0, reg1, reg2);
		printf ("LEA %s,[%s+1]\n", reg2, reg2);
		return true;
	}

	return false;
}

int
main (int argc, char **argv)
{
	if (argc != 2)
		return -1;
	FILE *f = fopen (argv[1], "rb");
	if (!f) {
		perror ("fopen");
		return -2;
	}

#define MAXLINELEN (4096)
	char line [MAXLINELEN];
	char comment [MAXLINELEN];
	int len;
	while (EOF != (len = readline (f, line, MAXLINELEN, comment, MAXLINELEN))) {
		if (len < 1) {
			puts (comment);
			continue;
		}

#define MAXWORDS (512)
		char *words [MAXWORDS];
		int count;
		if (0 < (count = getwords (line, len, words, MAXWORDS))) 
		{
			if (!convert_to_safe_form (words, count)) {

				// If no conversion was required or possible,
				// just print the line.
				//
				for (int i=0; i < count;i++) {
					char *str = words [i];
					printf ("%s ", str);
				}
			}
		}
		
		puts(comment);
	}

	if (!found_data_section) {
		print_only_allowed_ret ();
	}

	return 0;
}

